import React from 'react';

import { Layout } from './common/layout';

export const NotAttached = () => {
  return (
    <Layout>
      <div className="jp-SparkConnector-notattached">
        Open a notebook to connect to Apache Spark
      </div>
    </Layout>
  );
};
